import os
import re
from collections.abc import Iterator
from pathlib import Path
from typing import Any, Optional, AnyStr

from tutorial.util.request_meta import RequestMeta


def get_job_attachments(request_meta: RequestMeta, party_id: str, resource_id: str, solution_id: str,
                        job_ids: list[str]) -> Iterator[list[dict[str, Any]]]:
    """Retrieves the pages of the results calculations for a list of jobs. i.e., a list of attachments

    Arguments:
    party_id -- The party_id associated with the job
    solution_id -- Solution id as a string
    resource_id -- The resource_id associated with the job
    job_ids -- a list of job ids as strings
    """
    url = (f'{request_meta.base_url}/parties/{party_id}/models/{solution_id}/resource-types/Field/resources/{resource_id}'
           f'/insight-attachments')
    params = [('api-version', request_meta.fb_api_version)] + [('insightIds', job_id) for job_id in job_ids]

    response = request_meta.session.get(url, headers=request_meta.headers, params=params)
    response.raise_for_status()
    page = response.json()
    yield page['value']
    while 'nextLink' in page:
        skip_token_index = page['nextLink'].rfind("skipToken", )
        params = list(filter(lambda param: param[0] != 'skipToken', params))
        params.append(('skipToken', page['nextLink'][skip_token_index + len('skipToken='):]))
        response = request_meta.session.get(url, headers=request_meta.headers, params=params)
        response.raise_for_status()
        page = response.json()
        yield page['value']


def get_attachment_bytes(request_meta: RequestMeta, party_id: str, resource_id: str, solution_id: str,
                         attachment_id: str) -> bytes | None:
    """Retrieves a bytes attachment from the results of calculations.

    Arguments:
    party_id -- UUID of the party as a string
    resource_id -- UUID of the resource as a string
    solution_id -- solution id as a string
    attachment_id -- attachment id as a string
    """
    url = (f'{request_meta.base_url}/parties/{party_id}/models/{solution_id}/resource-types/Field/resources/{resource_id}'
           f'/insight-attachments/{attachment_id}/file')
    params = {'api-version': request_meta.fb_api_version}

    response = request_meta.session.get(url, headers=request_meta.headers, params=params)
    response.raise_for_status()
    return response.content


def save_attachments(request_meta: RequestMeta, output_path: Path, party_id: str, resource_id: str, solution_id: str,
                     job_id: str, attachment_pages: list[list[dict[str, Any]]],
                     job_id_regex: Optional[AnyStr] = None) -> None:
    """Retrieves all attachments.

    Arguments:
    output_path -- the root of the desire output path
    party_id -- UUID of the party as a string
    resource_id -- UUID of the resource as a string
    solution_id -- solution id as a string
    attachment_pages -- pages of JSON of all the attachments
    job_id_regex -- an optional regex that matches the job_ids to download attachments
    """
    attachments = []
    for attachment_page in attachment_pages:
        attachments = attachments + attachment_page

    filtered_attachments = attachments
    if job_id_regex is not None:
        regex = re.compile(job_id_regex)
        filtered_attachments = filter(lambda job: regex.match(job['insightId']), attachments)

    for attachment in filtered_attachments:
        attachment_id = attachment['id']
        attachment_bytes = get_attachment_bytes(request_meta, party_id, resource_id, solution_id, attachment_id)
        _, extension = os.path.splitext(attachment['originalFileName'])
        attachment_path = Path(os.path.join(output_path, f'{job_id}-{attachment_id}{extension}'))
        with attachment_path.open('wb') as binary_file:
            binary_file.write(attachment_bytes)
