# Copyright © 2024 AgPowered Services
import asyncio
import json
import logging
import os
import time
from dataclasses import dataclass, field
from datetime import date, timedelta
from enum import StrEnum
from pathlib import Path
from typing import Any, TextIO

import yaml
from requests.exceptions import HTTPError
from timeit import default_timer
from tutorial.adma.insight_attachment_util import get_job_attachments, save_attachments
from tutorial.adma.solution import submit_request, get_request_status
from tutorial.adma.jwt import get_jwt
from tutorial.util.headers import create_bearer_token_header
from tutorial.util.logging import configure_logging
from tutorial.util.request_meta import RequestMeta

logger = logging.getLogger('bulk-imagery')


class ImageResult(StrEnum):
    SUCCESS = 'Success'
    NO_SCENES_FOUND = 'No scenes found'
    TOO_CLOUDY = 'Too much cloud coverage'
    INCOMPLETE = 'Did not complete'
    UNKNOWN = 'Unknown'
    ERROR = 'Error'


@dataclass
class Work:
    job_id: str = field(default=None)
    solution_id: str = field(default=None)
    submit_body: dict[str, Any] = field(default=None)
    status_body: dict[str, Any] = field(default=None)
    image_date: date = field(default=None)

    @classmethod
    def create(cls, party_id: str, field_id: str, solution_id: str, image_date: date, layer: str):
        job_id = f'{image_date.isoformat()}-{layer}'

        submit_body = {
            'requestPath': f'/v1/imagery-add-on/{job_id}',
            'partnerRequestBody': {
                'partyId': party_id,
                'collection': 'sentinel-2-l2a',
                'layer': layer,
                'resourceId': field_id,
                'resourceType': 'Field',
                'sceneStartDateTime': f'{image_date.isoformat()}T00:00:00',
                'sceneEndDateTime': f'{image_date.isoformat()}T23:59:59'
            }
        }

        status_body = {
            'requestPath': f'/v1/imagery-add-on/{job_id}',
        }

        return cls(job_id, solution_id, submit_body, status_body, image_date)


async def generate_true_color(config: dict[str, Any], base_path: str, number_of_days: int, image_date: date,
                              party_id: str, field_id: str, solution_id: str) -> list[date]:
    layer = 'TRUE_COLOR'
    jwt = get_jwt(RequestMeta(), config)
    request_meta = RequestMeta.with_config(config, create_bearer_token_header(jwt))

    logger.info(f"Attempting to generate '{layer}' for {number_of_days} days starting with {image_date.isoformat()}.")

    coroutines = []
    for i in range(number_of_days):
        work = Work.create(party_id, field_id, solution_id, image_date, layer)
        coroutines.append(asyncio.to_thread(generate_image, request_meta, base_path, work))
        image_date = image_date + timedelta(days=1)

    results = await asyncio.gather(*coroutines)

    dates_with_imagery = []
    for image_result, work, poll_count, execution_time in results:
        if image_result == ImageResult.SUCCESS:
            dates_with_imagery.append(work.image_date)
            logger.info(f"Generated true color image on {work.image_date.isoformat()} with {work.job_id}, "
                        f"polled {poll_count} time(s), and completed execution in {execution_time}.")

        logger.info(f"Request completed for {work.job_id} result is '{image_result.value}'.")

    dates_for_output = [item.isoformat() for item in dates_with_imagery]
    logger.info(f'For the given range of dates, the following dates successfully downloaded {layer} imagery. '
                f'{dates_for_output}')
    return dates_with_imagery


async def generate_scouting(config: dict[str, Any], base_path: str, dates_with_imagery: list[date], party_id: str,
                            field_id: str, solution_id: str) -> list[date]:
    layer = 'SCOUTING_MAP'
    jwt = get_jwt(RequestMeta(), config)
    request_meta = RequestMeta.with_config(config, create_bearer_token_header(jwt))

    logger.info(f"Attempting to generate '{layer}' for days found with imagery.")

    coroutines = []
    for date_with_imagery in dates_with_imagery:
        work = Work.create(party_id, field_id, solution_id, date_with_imagery, layer)
        coroutines.append(asyncio.to_thread(generate_image, request_meta, base_path, work))

    results = await asyncio.gather(*coroutines)

    dates_with_non_cloudy_imagery = []
    for image_result, work, poll_count, execution_time in results:
        if image_result == ImageResult.SUCCESS:
            dates_with_non_cloudy_imagery.append(work.image_date)
            logger.info(f"Generated scouting image on {work.image_date.isoformat()} with {work.job_id}, "
                        f"polled {poll_count} time(s), and completed execution in {execution_time}.")

        logger.info(f"Request completed for {work.job_id} result is '{image_result.value}'.")

    dates_for_output = [item.isoformat() for item in dates_with_non_cloudy_imagery]
    logger.info(f'For the given range of dates, the following dates successfully downloaded non-cloudy {layer} '
                f'imagery. {dates_for_output}')
    return dates_with_non_cloudy_imagery


async def generate_vegetation(config: dict[str, Any], base_path: str, dates_with_non_cloudy_imagery: list[date],
                              party_id: str, field_id: str, solution_id: str) -> None:
    layer = 'VEGETATION_MAP'
    jwt = get_jwt(RequestMeta(), config)
    request_meta = RequestMeta.with_config(config, create_bearer_token_header(jwt))

    logger.info(f"Attempting to generate '{layer}' for days found with non-cloudy imagery.")

    coroutines = []
    for date_with_non_cloudy_imagery in dates_with_non_cloudy_imagery:
        work = Work.create(party_id, field_id, solution_id, date_with_non_cloudy_imagery, layer)
        coroutines.append(asyncio.to_thread(generate_image, request_meta, base_path, work))

    results = await asyncio.gather(*coroutines)

    for image_result, work, poll_count, execution_time in results:
        if image_result == ImageResult.SUCCESS:
            logger.info(f"Generated vegetation image on {work.image_date.isoformat()} with {work.job_id}, "
                        f"polled {poll_count} time(s), and completed execution in {execution_time}.")

        logger.info(f"Request completed for {work.job_id} result is '{image_result.value}'.")


def poll_job(request_meta: RequestMeta, work: Work, log_file: TextIO) -> tuple[ImageResult, int]:
    status_check_count = 1
    imagery_status_response = get_request_status(request_meta, work.solution_id, work.status_body)
    while (imagery_status_response['status'] == 'RUNNING'
           or imagery_status_response['status'] == 'WAITING') and status_check_count <= 50:
        status_check_count = status_check_count + 1
        time.sleep(1)
        imagery_status_response = get_request_status(request_meta, work.solution_id, work.status_body)
    match imagery_status_response['status']:
        case 'SUCCEEDED':
            log_file.write(json.dumps(imagery_status_response, indent=2))
            log_file.write('\n')
            result = ImageResult.SUCCESS
        case 'FAILED':
            log_file.write(json.dumps(imagery_status_response, indent=2))
            log_file.write('\n')
            if '04019' in imagery_status_response['message']:
                result = ImageResult.NO_SCENES_FOUND
            elif '06001' in imagery_status_response['message']:
                result = ImageResult.TOO_CLOUDY
            else:
                result = ImageResult.UNKNOWN
        case 'RUNNING' | 'WAITING':
            log_file.write('Image generation failed to finish in requested timeframe.')
            log_file.write('\n')
            result = ImageResult.INCOMPLETE
        case _:
            log_file.write(f'Unknown status {imagery_status_response["status"]}')
            log_file.write('\n')
            result = ImageResult.UNKNOWN
    return result, status_check_count


def generate_image(request_meta: RequestMeta, base_path: str, work: Work) -> tuple[ImageResult, Work, int, timedelta]:
    party_id = work.submit_body['partnerRequestBody']['partyId']
    resource_id = work.submit_body['partnerRequestBody']['resourceId']

    poll_count = 0
    t = default_timer()
    path = Path(os.path.join(base_path, party_id, resource_id, work.solution_id))
    if not os.path.exists(path):
        os.makedirs(path)

    log = Path(os.path.join(path, f'{work.job_id}.txt'))
    with log.open('w') as log_file:
        log_file.write(json.dumps(work.submit_body, indent=2))
        log_file.write('\n')

        try:
            imagery_response = submit_request(request_meta, work.solution_id, work.submit_body)
            log_file.write(json.dumps(imagery_response, indent=2))
            log_file.write('\n')

            image_result, poll_count = poll_job(request_meta, work, log_file)

            if image_result is ImageResult.SUCCESS:
                attachment_responses = get_job_attachments(request_meta, party_id, resource_id, work.solution_id,
                                                           [work.job_id])
                attachment_pages = list(attachment_responses)
                save_attachments(request_meta, path, party_id, resource_id, work.solution_id, work.job_id,
                                 attachment_pages)

            return image_result, work, poll_count, timedelta(seconds=default_timer() - t)
        except HTTPError as ex:
            log_file.write(f'Error for {work.job_id} {ex.response.text} {json.dumps(work.submit_body, indent=2)}')
            log_file.write('\n')
            return ImageResult.ERROR, work, poll_count, timedelta(seconds=default_timer() - t)


async def main():
    config_resource = './resource/application_config.yml'
    logging_resource = './resource/logging.yml'

    with open(logging_resource) as logging_file:
        configure_logging(yaml.safe_load(logging_file))

    logger.info(f"Logging configured using '{logging_resource}'")

    # layers = ['TRUE_COLOR', 'SCOUTING_MAP', 'VEGETATION_MAP']
    number_of_days = 31
    image_date = date.fromisoformat('2023-08-01')
    party_id = 'REPLACE_ME'  # party_id used in create_party
    field_id = 'REPLACE_ME'  # field_id used in create_field

    with open(config_resource) as config_file:
        config = yaml.safe_load(config_file)

    logger.info(f"Application configured using '{config_resource}'")
    base_path = './output'
    solution_id = config['imagery_solution_id']

    t = default_timer()
    # First, attempt to generate true color imagery as this is not affected by cloudiness.
    # As a side-effect this will download all available true color imagery for the date range.
    # Another side-effect is that this will ingest most of the bands needed for image generation of other layers.
    dates_with_imagery = await generate_true_color(config, base_path, number_of_days, image_date, party_id, field_id,
                                                   solution_id)

    # Second, attempt generation of scouting imagery as we now know scenes are available for these days.
    # Most of the necessary bands will have been ingested and cached by this time.
    # We are filtering the good dates down with each pass to help avoid unnecessary computation and cost.
    dates_with_non_cloudy_imagery = await generate_scouting(config, base_path, dates_with_imagery, party_id, field_id,
                                                            solution_id)

    # Finally, attempt generation of vegetation imagery as we now know non-cloudy scenes are available for these days.
    # Since only dates with 80% or less clouds are used, we don't attempt to generate any imagery that would result in
    # a failure due to cloudiness.
    # We are filtering the good dates down with each pass to help avoid unnecessary computation and cost.
    await generate_vegetation(config, base_path, dates_with_non_cloudy_imagery, party_id, field_id, solution_id)
    logger.info(f'Accumulated time = {timedelta(seconds=default_timer() - t)} '
                f'to generate all image layer for a field.')


if __name__ == '__main__':
    asyncio.run(main=main())
