import asyncio
import json
import logging
import os
import time
from dataclasses import dataclass, field, asdict
from datetime import date, timedelta
from pathlib import Path
from typing import List, Any

import yaml

from requests.exceptions import HTTPError
from timeit import default_timer
from tutorial.adma.insight import get_insight
from tutorial.adma.jwt import get_jwt
from tutorial.adma.solution import submit_request, get_request_status
from tutorial.util.headers import create_bearer_token_header
from tutorial.util.logging import configure_logging
from tutorial.util.request_meta import RequestMeta

logger = logging.getLogger('async-bulk-gdd')


@dataclass(order=True)
class GddMeasurement:
    date: str = field(default=None)
    daily_contribution: float = field(default=0.0, compare=False)
    cumulative_sum: float = field(default=0.0, compare=False)
    success: bool = field(default=False, compare=False)


@dataclass
class GddData:
    temperature_unit: str = field(default='C')
    gdd_measurements: List[GddMeasurement] = field(default_factory=lambda: [])


@dataclass
class Work:
    job_id: str = field(default=None)
    solution_id: str = field(default=None)
    submit_body: dict[str, Any] = field(default=None)
    status_body: dict[str, Any] = field(default=None)
    gdd_date: date = field(default=None)

    @classmethod
    def create(cls, party_id: str, field_id: str, solution_id: str, gdd_date: date, crop: str, temperature_unit: str):
        job_id = f'{gdd_date.isoformat()}-{crop}'

        submit_body = {
            'requestPath': f'/v1/gdd/{job_id}',
            'partnerRequestBody': {
                'partyId': party_id,
                'crop': crop,
                'resourceId': field_id,
                'resourceType': 'Field',
                'gddStartDate': gdd_date.isoformat(),
                'gddEndDate': gdd_date.isoformat(),
                'temperatureUnit': temperature_unit
            }
        }

        status_body = {
            'requestPath': f'/v1/gdd/{job_id}',
        }

        return cls(job_id, solution_id, submit_body, status_body, gdd_date)


async def generate_gdd_data(config: dict[str, Any], base_path: str, number_of_days: int, start_date: date, crop: str,
                            party_id: str, field_id: str, solution_id: str) -> None:
    path = Path(os.path.join(base_path, party_id, field_id, solution_id))
    if not os.path.exists(path):
        os.makedirs(path)

    jwt = get_jwt(RequestMeta(), config)
    request_meta = RequestMeta.with_config(config, create_bearer_token_header(jwt))

    logger.info(f"Attempting to generate '{crop}' GDD for {number_of_days} days starting with "
                f"{start_date.isoformat()}.")

    temperature_unit = 'C'
    coroutines = []
    working_date = start_date
    for i in range(number_of_days):
        work = Work.create(party_id, field_id, solution_id, working_date, crop, temperature_unit)
        coroutines.append(asyncio.to_thread(calculate_gdd, request_meta, path, work))
        working_date = working_date + timedelta(days=1)

    results = await asyncio.gather(*coroutines)

    failed_days = []
    cumulative_sum = 0.0
    gdd_data = GddData(temperature_unit)
    for gdd_measurement, job_id, status_check_count, execution_time in results:
        if gdd_measurement.success:
            cumulative_sum = cumulative_sum + gdd_measurement.daily_contribution
            gdd_measurement.cumulative_sum = cumulative_sum
            gdd_data.gdd_measurements.append(gdd_measurement)
            logger.info(f"Calculated GDD for '{crop}' on {gdd_measurement.date} with {job_id}, "
                        f"checked status {status_check_count} time(s), and completed execution in {execution_time}.")
        else:
            failed_days.append(gdd_measurement.date)
            logger.info(f"Unable to calculate GDD for '{crop}' on {gdd_measurement.date}.")

    if failed_days:
        logger.info(f"The following days {failed_days} did not generate '{crop}' GDD.")

    logger.info(f"Generated '{crop}' GDD for {number_of_days} days starting with {start_date.isoformat()}.")

    end_date = start_date + timedelta(days=(number_of_days - 1))
    output_path = Path(os.path.join(path, f'{start_date.isoformat()}-{end_date.isoformat()}-{crop}.json'))
    with open(output_path, "w") as outfile:
        json.dump(asdict(gdd_data), outfile, indent=2)


def calculate_gdd(request_meta: RequestMeta, path: Path, work: Work) -> tuple[GddMeasurement, str, int, timedelta]:
    party_id = work.submit_body['partnerRequestBody']['partyId']
    resource_id = work.submit_body['partnerRequestBody']['resourceId']
    resource_type = work.submit_body['partnerRequestBody']['resourceType']
    gdd_measurement = GddMeasurement(work.gdd_date.isoformat())

    t = default_timer()
    log = Path(os.path.join(path, f'{work.job_id}.txt'))
    with log.open('w') as log_file:
        log_file.write(json.dumps(work.submit_body, indent=2))
        log_file.write('\n')

        try:
            gdd_response = submit_request(request_meta, work.solution_id, work.submit_body)
            log_file.write(json.dumps(gdd_response, indent=2))
            log_file.write('\n')

            status_check_count = 1
            gdd_status_response = get_request_status(request_meta, work.solution_id, work.status_body)
            while (gdd_status_response['status'] == 'RUNNING'
                   or gdd_status_response['status'] == 'WAITING') and status_check_count <= 25:
                time.sleep(1)
                status_check_count = status_check_count + 1
                gdd_status_response = get_request_status(request_meta, work.solution_id, work.status_body)
            match gdd_status_response['status']:
                case 'SUCCEEDED':
                    log_file.write(json.dumps(gdd_status_response, indent=2))
                    log_file.write('\n')

                    insight_response = get_insight(request_meta, party_id, work.solution_id, resource_type,
                                                   resource_id, work.job_id)
                    daily_contribution = insight_response['measurements']['computedGdd']['value']
                    gdd_measurement = GddMeasurement(work.gdd_date.isoformat(), daily_contribution, 0.0, True)

                    log_file.write(json.dumps(insight_response, indent=2))
                    log_file.write('\n')
                case 'FAILED':
                    log_file.write(json.dumps(gdd_status_response, indent=2))
                    log_file.write('\n')
                case 'RUNNING' | 'WAITING':
                    log_file.write('GDD calculation failed to finish in requested timeframe.')
                    log_file.write('\n')
                case _:
                    log_file.write(f'Unknown status {gdd_status_response["status"]}')
                    log_file.write('\n')
        except HTTPError as ex:
            log_file.write(f'Error for {work.job_id} {ex.response.text} {json.dumps(work.submit_body, indent=2)}')
            log_file.write('\n')
    return gdd_measurement, work.job_id, status_check_count, timedelta(seconds=default_timer() - t)


async def main():
    config_resource = './resource/application_config.yml'
    logging_resource = './resource/logging.yml'

    with open(logging_resource) as logging_file:
        configure_logging(yaml.safe_load(logging_file))

    logger.info(f"Logging configured using '{logging_resource}'")

    number_of_days = 31
    start_date = date.fromisoformat('2023-08-01')
    crop = 'canola'  # ['canola', 'corn', 'cotton', 'sugar_beet', 'soybean', 'wheat']
    party_id = 'REPLACE_ME'  # party_id used in create_party
    field_id = 'REPLACE_ME'  # field_id used in create_field

    with open(config_resource) as config_file:
        config = yaml.safe_load(config_file)

    logger.info(f"Application configured using '{config_resource}'")

    base_path = './output'
    solution_id = config['gdd_solution_id']

    t = default_timer()
    await generate_gdd_data(config, base_path, number_of_days, start_date, crop, party_id, field_id, solution_id)
    logger.info(f'Accumulated time = {timedelta(seconds=default_timer() - t)} '
                f'to generate GDD for a field.')


if __name__ == '__main__':
    asyncio.run(main=main())