# Copyright © 2024 AgPowered Services
import json
import logging
import os
import time
from datetime import date, datetime
from pathlib import Path
from typing import Any

import yaml
from requests.exceptions import HTTPError

from tutorial.adma.insight_attachment_util import get_job_attachments, save_attachments
from tutorial.adma.solution import submit_request, get_request_status
from tutorial.adma.jwt import get_jwt
from tutorial.util.headers import create_bearer_token_header
from tutorial.util.logging import configure_logging
from tutorial.util.request_meta import RequestMeta


def handle_job(request_meta: RequestMeta, base_path: str, job_id: str, solution_id: str, submit_body: dict[str, Any],
               status_body: dict[str, Any]) -> None:
    party_id = submit_body['partnerRequestBody']['partyId']
    resource_id = submit_body['partnerRequestBody']['resourceId']

    path = Path(os.path.join(base_path, party_id, resource_id, solution_id))
    if not os.path.exists(path):
        os.makedirs(path)

    log = Path(os.path.join(path, f'{job_id}.txt'))
    with log.open('w') as log_file:
        log_file.write(json.dumps(submit_body, indent=2))
        log_file.write('\n')

        try:
            job_submit_response = submit_request(request_meta, solution_id, submit_body)
            log_file.write(json.dumps(job_submit_response, indent=2))
            log_file.write('\n')

            status_check_count = 1
            job_status_response = get_request_status(request_meta, solution_id, status_body)
            while (job_status_response['status'] == 'RUNNING'
                   or job_status_response['status'] == 'WAITING') and status_check_count <= 25:
                status_check_count = status_check_count + 1
                time.sleep(1)
                job_status_response = get_request_status(request_meta, solution_id, status_body)
            match job_status_response['status']:
                case 'SUCCEEDED':
                    log_file.write(json.dumps(job_status_response, indent=2))
                    log_file.write('\n')

                    attachment_responses = get_job_attachments(request_meta, party_id, resource_id, solution_id,                                                               [job_id])
                    attachment_pages = list(attachment_responses)
                    save_attachments(request_meta, path, party_id, resource_id, solution_id, job_id, attachment_pages)
                case 'FAILED':
                    log_file.write(json.dumps(job_status_response, indent=2))
                    log_file.write('\n')
                case 'RUNNING' | 'WAITING':
                    log_file.write('Job failed to finish in requested timeframe.')
                    log_file.write('\n')
                case _:
                    log_file.write(f'Unknown status {job_status_response["status"]}')
                    log_file.write('\n')
        except HTTPError as ex:
            log_file.write(f'Error for {job_id} {ex.response.text} {json.dumps(submit_body, indent=2)}')
            log_file.write('\n')


def main():
    config_resource = './resource/application_config.yml'
    logging_resource = './resource/logging.yml'

    with open(logging_resource) as logging_file:
        configure_logging(yaml.safe_load(logging_file))

    logger = logging.getLogger('crop-id')

    logger.info(f"Logging configured using '{logging_resource}'")

    party_id = 'REPLACE_ME'  # party_id used in create_party
    field_id = 'REPLACE_ME'  # field_id used in create_field

    job_id = f'crop-id_{datetime.today().strftime("%Y-%m-%dT%H:%M:%S%z") }'.replace(':', '_')

    with open(config_resource) as config_file:
        config = yaml.safe_load(config_file)

    logger.info(f"Application configured using '{config_resource}'")
    solution_id = config['cropid_solution_id']
    job_request_body = {
        'requestPath': f'/v1/crop-id/{job_id}',
        'partnerRequestBody': {
            'partyId': party_id,
            'resourceId': field_id,
            'resourceType': 'Field',
            'startYear': '2022',
            'endYear': '2023'
        }
    }

    job_status_body = {
        'requestPath': f'/v1/crop-id/{job_id}',
    }

    jwt = get_jwt(RequestMeta(), config)
    request_meta = RequestMeta.with_config(config, create_bearer_token_header(jwt))
    handle_job(request_meta, './output', job_id, solution_id, job_request_body, job_status_body)

    logger.info(f"Completed crop-id request for {job_id}.")


if __name__ == '__main__':
    main()
