# Copyright © 2024 AgPowered Services
import json
import logging
import os
import time
from datetime import date
from pathlib import Path
from typing import Any

import yaml
from requests.exceptions import HTTPError

from tutorial.adma.insight_attachment_util import get_job_attachments, save_attachments
from tutorial.adma.solution import submit_request, get_request_status
from tutorial.adma.jwt import get_jwt
from tutorial.util.headers import create_bearer_token_header
from tutorial.util.logging import configure_logging
from tutorial.util.request_meta import RequestMeta


def generate_image(request_meta: RequestMeta, base_path: str, job_id: str, solution_id: str, submit_body: dict[str, Any],
                   status_body: dict[str, Any]) -> None:
    party_id = submit_body['partnerRequestBody']['partyId']
    resource_id = submit_body['partnerRequestBody']['resourceId']

    path = Path(os.path.join(base_path, party_id, resource_id, solution_id))
    if not os.path.exists(path):
        os.makedirs(path)

    log = Path(os.path.join(path, f'{job_id}.txt'))
    with log.open('w') as log_file:
        log_file.write(json.dumps(submit_body, indent=2))
        log_file.write('\n')

        try:
            imagery_response = submit_request(request_meta, solution_id, submit_body)
            log_file.write(json.dumps(imagery_response, indent=2))
            log_file.write('\n')

            status_check_count = 1
            imagery_status_response = get_request_status(request_meta, solution_id, status_body)
            while (imagery_status_response['status'] == 'RUNNING'
                   or imagery_status_response['status'] == 'WAITING') and status_check_count <= 25:
                status_check_count = status_check_count + 1
                time.sleep(1)
                imagery_status_response = get_request_status(request_meta, solution_id, status_body)
            match imagery_status_response['status']:
                case 'SUCCEEDED':
                    log_file.write(json.dumps(imagery_status_response, indent=2))
                    log_file.write('\n')

                    attachment_responses = get_job_attachments(request_meta, party_id, resource_id, solution_id,                                                               [job_id])
                    attachment_pages = list(attachment_responses)
                    save_attachments(request_meta, path, party_id, resource_id, solution_id, job_id, attachment_pages)
                case 'FAILED':
                    log_file.write(json.dumps(imagery_status_response, indent=2))
                    log_file.write('\n')
                case 'RUNNING' | 'WAITING':
                    log_file.write('Image generation failed to finish in requested timeframe.')
                    log_file.write('\n')
                case _:
                    log_file.write(f'Unknown status {imagery_status_response["status"]}')
                    log_file.write('\n')
        except HTTPError as ex:
            log_file.write(f'Error caught for {job_id} : {str(ex)}\nJSON request body:'
                           f'\n{json.dumps(submit_body, indent=2)}\n')



def main():
    config_resource = './resource/application_config.yml'
    logging_resource = './resource/logging.yml'

    with open(logging_resource) as logging_file:
        configure_logging(yaml.safe_load(logging_file))

    logger = logging.getLogger('imagery')

    logger.info(f"Logging configured using '{logging_resource}'")

    layer = 'VEGETATION_MAP'  # ['SCOUTING_MAP', 'VEGETATION_MAP', 'TRUE_COLOR']
    image_date = date.fromisoformat('2023-08-28')
    party_id = 'REPLACE_ME'  # party_id used in create_party
    field_id = 'REPLACE_ME'  # field_id used in create_field
    job_id = f'{image_date.isoformat()}-{layer}'

    with open(config_resource) as config_file:
        config = yaml.safe_load(config_file)

    logger.info(f"Application configured using '{config_resource}'")
    solution_id = config['imagery_solution_id']

    imagery_submit_body = {
        'requestPath': f'/v1/imagery-add-on/{job_id}',
        'partnerRequestBody': {
            'partyId': party_id,
            'collection': 'sentinel-2-l2a',
            'layer': layer,
            'resourceId': field_id,
            'resourceType': 'Field',
            'sceneStartDateTime': f'{image_date.isoformat()}T00:00:00',
            'sceneEndDateTime': f'{image_date.isoformat()}T23:59:59'
        }
    }

    imagery_status_body = {
        'requestPath': f'/v1/imagery-add-on/{job_id}',
    }

    jwt = get_jwt(RequestMeta(), config)
    request_meta = RequestMeta.with_config(config, create_bearer_token_header(jwt))
    generate_image(request_meta, './output', job_id, solution_id, imagery_submit_body, imagery_status_body)

    logger.info(f"Completed imagery request of {layer} for {job_id}.")


if __name__ == '__main__':
    main()
